/* $Id: grid.cpp,v 1.4 2006/01/18 12:00:11 cvsroot Exp $ */

/* GenLib2 - General Library 2
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#define DB_WARNING 0
#include <cassert>
#include <ostream>
#include <cfloat>
#include <cmath>
#include "ex.h"
#include "grid.h"
#include "osform.h"

namespace GenLib2
{

// constructor

Grid::Grid()
{
	DB_ENTER("Grid::constr");

	int i, j;

	z = new node;
	assert(z != 0);
	for (i = 0; i < MAXGRID; i++)
		for (j = 0; j < MAXGRID; j++)
			grid[i][j] = z;

	origin.x = origin.y = 0.;
	origin.c = 0;
	width = 1.;
}


// destructor

Grid::~Grid()
{
    DB_ENTER("Grid::destr");

	clear();
}


// clear grid

void Grid::clear()
{
    DB_ENTER("Grid::clear");

    int i, j;
	struct node *t, *p;

	for (i = 0; i < MAXGRID; i++)
		for (j = 0; j < MAXGRID; j++)
			{
			t = grid[i][j];
			while (t != z)
				{
				p = t;
				t = t->next;
				delete p;
				}
			grid[i][j] = z;
			}
}


// set gird resolution

void Grid::set(struct range x, struct range y)
{
    DB_ENTER("Grid::set");

	double dx = x.max - x.min;
	double dy = y.max - y.min;

	double wx = dx / MAXGRID;
	double wy = dy / MAXGRID;

	width = wx > wy ? wx : wy;

	origin.x = x.min;
	origin.y = y.min;
	origin.c = 0;
}


// print info

void Grid::info(ostream &os) const
{
	DB_ENTER("Grid::info");

	os << "Grid range\n";
	os << "  in x: " << origin.x << " - "
		<< origin.x + MAXGRID * width << endl;
	os << "  in y: " << origin.y << " - "
		<< origin.y + MAXGRID * width << endl;
	os << "Width of grid square : " << width << endl;
	os << "Origin : " << origin.x << " - " << origin.y << endl;
}


// insert point in grid

void Grid::insert(struct point p)
{
    DB_ENTER("Grid::insert");

	struct node *t = new node;
	assert(t != 0);

	int i = int((p.x - origin.x) / width);
	int j = int((p.y - origin.y) / width);

	if (i < 0) i = 0;
	if (j < 0) j = 0;
	if (i >= MAXGRID) i = MAXGRID - 1;
	if (j >= MAXGRID) j = MAXGRID - 1;

	t->p = p; t->next = grid[i][j];
	grid[i][j] = t;
}


// count number of points

size_t Grid::size() const
{
	DB_ENTER("Grid::size");

	int i, j;
	size_t n = 0;
	struct node *t;

	for (i = 0; i < MAXGRID; i++)
		for (j = 0; j < MAXGRID; j++)
			for (t = grid[i][j]; t != z; t = t->next)
				n++;

    return n;
}


// print number of points per grid square

void Grid::occupied(ostream& os) const
{
	DB_ENTER("Grid::occupied");

	int i, j;
	struct node* t;

	for (i = 0; i < MAXGRID; i++)
		{
		os << "row " << i << ": ";
		for (j = 0; j < MAXGRID; j++)
			{
			int n = 0;
			for (t = grid[i][j]; t != z; t = t->next)
				n++;
			os << n << " ";
			}
		os << endl;
		}
}


// calculate average inside the given rectangle

int Grid::average(struct rect range, struct point& o) const
{
	DB_ENTER("Grid::average");

	register struct node *t;
	register int i, j;
	double dx, dy, s, p;
	double sum_pc = 0.;
	double sum_p = 0.;
	int count = 0;

	int i1 = int((range.x1 - origin.x) / width);
	if (i1 < 0) i1 = 0;
	int i2 = int((range.x2 - origin.x) / width);
	if (i2 >= MAXGRID) i2 = MAXGRID - 1;
	int j1 = int((range.y1 - origin.y) / width);
	if (j1 < 0) j1 = 0;
	int j2 = int((range.y2 - origin.y) / width);
	if (j2 >= MAXGRID) j2 = MAXGRID - 1;

	for (i = i1; i <= i2; i++)
		for (j = j1; j <= j2; j++)
			for (t = grid[i][j]; t != z; t = t->next)
				if (insiderect(t->p, range))
					{
					dx = t->p.x - o.x;
					dy = t->p.y - o.y;
					s = fabs(dx) + fabs(dy);
					if (s < DBL_EPSILON) s = DBL_EPSILON;
					p = 1. / (s * s);
					sum_pc += p * t->p.c;
					sum_p += p;
					count++;
					}

	if (sum_p > 0.)
		o.c = int(sum_pc / sum_p);

	return count;
}


// find closest point

struct point* Grid::closest(struct rect range, struct point& o) const
{
	DB_ENTER("Grid::closest");

	register struct node *t;
	register int i, j;
	struct point* c = 0;
	double dx, dy, ssq, ssq_min = DBL_MAX;

	int i1 = int((range.x1 - origin.x) / width);
	if (i1 < 0) i1 = 0;
	int i2 = int((range.x2 - origin.x) / width);
	if (i2 >= MAXGRID) i2 = MAXGRID - 1;
	int j1 = int((range.y1 - origin.y) / width);
	if (j1 < 0) j1 = 0;
	int j2 = int((range.y2 - origin.y) / width);
	if (j2 >= MAXGRID) j2 = MAXGRID - 1;

	for (i = i1; i <= i2; i++)
		for (j = j1; j <= j2; j++)
			for (t = grid[i][j]; t != z; t = t->next)
				if (insiderect(t->p, range))
					{
					dx = t->p.x - o.x;
					dy = t->p.y - o.y;
					ssq = dx * dx + dy * dy;
					if (ssq < ssq_min)
						{
						ssq_min = ssq;
						c = &(t->p);
						}
					}

	return c;
}


// find bonding box

bool Grid::bbox(struct range& rx, struct range& ry) const
{
	DB_ENTER("Grid::box");

	int i, j;
	struct node *t;

	for (i = 0; i < MAXGRID; i++)
		for (j = 0; j < MAXGRID; j++)
			for (t = grid[i][j]; t != z; t = t->next)
				{
				rx.min = rx.max = t->p.x;
				ry.min = ry.max = t->p.y;
				goto escape;
				}

escape:
	for (i = 0; i < MAXGRID; i++)
		for (j = 0; j < MAXGRID; j++)
			for (t = grid[i][j]; t != z; t = t->next)
				{
				if (t->p.x < rx.min) rx.min = t->p.x;
				if (t->p.x > rx.max) rx.max = t->p.x;
				if (t->p.y < ry.min) ry.min = t->p.y;
				if (t->p.y > ry.max) ry.max = t->p.y;
				}

	return true;
}


// print grid points

void Grid::print(ostream& os) const
{
	DB_ENTER("Grid::print");

	int i, j;
	struct node *t;

	for (i = 0; i < MAXGRID; i++)
		for (j = 0; j < MAXGRID; j++)
			for (t = grid[i][j]; t != z; t = t->next)
				os << t->p.x << " " << t->p.y << " " << t->p.c << endl;
}


// print grid points using format string

void Grid::print(ostream& os, char* fmt) const
{
	DB_ENTER("Grid::print formated");

	int i, j;
	struct node *t;

	for (i = 0; i < MAXGRID; i++)
		for (j = 0; j < MAXGRID; j++)
			for (t = grid[i][j]; t != z; t = t->next)
				os << form(fmt, t->p.x, t->p.y, t->p.c);
}


// compute distance and check if less than min

inline void Grid::check(struct point p1, struct point p2)
{
	double d2;

	d2 = (p2.x - p1.x) * (p2.x - p1.x) + (p2.y - p1.y) * (p2.y - p1.y);
	if (d2 < min2)
		{
		min2 = d2;
		cp1 = p1;
		cp2 = p2;
		}
}


// check if inside rectangle

inline int Grid::insiderect(struct point p, struct rect range) const
{
	if (p.x < range.x1 || p.x > range.x2
	 || p.y < range.y1 || p.y > range.y2)
		return 0;
	else
		return 1;
}

} // namespace GenLib2
