/* $Id: vector2.h,v 1.2 2006/01/18 12:05:27 cvsroot Exp $ */

/* GenLib2 - General Library 2
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef GENLIB2_VECTOR2_INCLUDED
#define GENLIB2_VECTOR2_INCLUDED

#include <iostream>
#include "matrix.h"
//#include "math.h" // use own include file!

using namespace std;

namespace GenLib2
{

template<class T>
class Vector2
{
public:
	// Constructors, destructor, assignment operator
	Vector2() { x=0; y=0; }
	Vector2(T s) { init(s); }
	Vector2(T x, T y) { init(x,y); }
	Vector2(const Vector2& that) { init(that.x,that.y); }

	Vector2& init(T X, T Y) { x=X; y=Y; return *this; }
	Vector2& init(T s) { x=s; y=s; return *this; }

    Vector2& operator=(const Vector2& a) {
		return init((x=a.x),(y=a.y));
	}

    // composite assignment operators...
	/*
	Vector2& operator*=(T s) {
		x *= s; y *= s;
		return *this;
	}
	*/

    Vector2& operator+=(const Vector2& a) {
		x += a.x; y += a.y;
		return *this;
	}

    Vector2& operator-=(const Vector2& a) {
		x -= a.x; y -= a.y;
		return *this;
	}

    Vector2& operator*=(const Vector2& a) {
		x *= a.x; y *= a.y;
		return *this;
	}

    Vector2& operator/=(const Vector2& a) {
		x /= a.x; y /= a.y;
		return *this;
	}

	// matrix-vector products...
	Vector2& prod(const Matrix<T>& A, const Vector2& b) {
		assert(A.rows() == 2);
		assert(A.cols() == 2);

		x = A(0,0) * b.x + A(0,1) * b.y;
		y = A(1,0) * b.x + A(1,1) * b.y;

		return *this;
	}

	Vector2& prod_trn(const Matrix<T>& A, const Vector2& b) {
		assert(A.rows() == 2);
		assert(A.cols() == 2);

		x = A(0,0) * b.x + A(1,0) * b.y;
		y = A(0,1) * b.x + A(1,1) * b.y;

		return *this;
	}
	

	// Data members
	T x, y;
};


// non-member implementations...

// arithmetic operators...

template<class T> inline
Vector2<T> operator+(const Vector2<T>& a, const Vector2<T>& b)
{
	Vector2<T> tmp(a.x + b.x, a.y + b.y);
	return tmp;
}

template<class T> inline
Vector2<T> operator-(const Vector2<T>& a, const Vector2<T>& b)
{
	Vector2<T> tmp(a.x - b.x, a.y - b.y);
	return tmp;
}

template<class T> inline
Vector2<T> operator*(const Vector2<T>& a, const Vector2<T>& b)
{
	Vector2<T> tmp(a.x * b.x, a.y * b.y);
	return tmp;
}

template<class T> inline
Vector2<T> operator/(const Vector2<T>& a, const Vector2<T>& b)
{
	Vector2<T> tmp(a.x / b.x, a.y / b.y);
	return tmp;
}


// scalar operators...

template<class T> inline
Vector2<T> operator*(const Vector2<T>& a, T s)
{
	Vector2<T> tmp(a.x * s, a.y * s);
	return tmp;
}

template<class T> inline
Vector2<T> operator/(const Vector2<T>& a, T s)
{
	Vector2<T> tmp(a.x / s, a.y / s);
	return tmp;
}


// comparison operators...

template<class T> inline
bool operator==(const Vector2<T>& a, const Vector2<T>& b)
{
  return (a.x == b.x && a.y == b.y);
}

template<class T> inline
bool operator!=(const Vector2<T>& a, const Vector2<T>& b)
{
  return (a.x != b.x || a.y != b.y);
}


// input operator, input could be x y or (x,y)

template<class T> inline
istream& operator>>(istream& is, Vector2<T>& a)
{
	T x = 0, y = 0;
	char  c = 0;

	is >> c;
	if (c == '(')
		{
		is >> x >> c;
		if (c == ',') is >> y >> c;
		if (c != ')') is.clear(ios::badbit);
		}
	else
		{
		is.putback(c);
		is >> x >> y;
		}

	if (is) a = Vector2<T>(x, y);
	return is;
}


// output operator

template<class T> inline
ostream& operator<<(ostream& os, const Vector2<T>& a)
{
	return os << "( " << a.x << ", " << a.y << ")";
}


// mathematical functions...

// absolute value

template<class T> inline
T abs2(const Vector2<T>& a)
{
	return (a.x * a.x + a.y * a.y);
}

template<class T> inline
T abs(const Vector2<T>& a)
{
	return sqrt(abs2(a));
}


// dot product

template<class T> inline
T dot(const Vector2<T>& a, const Vector2<T>& b)
{
	return (a.x * b.x + a.y * b.y);
}


// matrix-vector multiplication operator

template<class T>
Vector2<T> operator*(const Matrix<T>& A, const Vector2<T>& b)
{
	assert(A.rows() == 2);
	assert(A.cols() == 2);
	
	Vector2<T> c;
	c.prod(A, b);
	//c.x = A(0,0) * b.x + A(0,1) * b.y;
	//c.y = A(1,0) * b.x + A(1,1) * b.y;

    return c;
}

} // namespace GenLib2

#endif /* GENLIB2_VECTOR2_INCLUDED */

